/*_############################################################################
  _## 
  _##  SNMP4J-UNIX-TRANSPORT - UnixSocketStreamTransportMapping.java  
  _## 
  _##  Copyright (C) 2021-2022  Frank Fock (SNMP4J.org)
  _##  
  _##  Licensed under the Apache License, Version 2.0 (the "License");
  _##  you may not use this file except in compliance with the License.
  _##  You may obtain a copy of the License at
  _##  
  _##      http://www.apache.org/licenses/LICENSE-2.0
  _##  
  _##  Unless required by applicable law or agreed to in writing, software
  _##  distributed under the License is distributed on an "AS IS" BASIS,
  _##  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  _##  See the License for the specific language governing permissions and
  _##  limitations under the License.
  _##  
  _##########################################################################*/
package org.snmp4j.transport.unix;

import org.snmp4j.SNMP4JSettings;
import org.snmp4j.TransportStateReference;
import org.snmp4j.log.LogAdapter;
import org.snmp4j.log.LogFactory;
import org.snmp4j.security.SecurityLevel;
import org.snmp4j.smi.Address;
import org.snmp4j.transport.*;

import java.io.IOException;
import java.io.UncheckedIOException;
import java.net.*;
import java.nio.ByteBuffer;
import java.nio.channels.*;
import java.nio.file.Files;
import java.util.Optional;
import java.util.Set;

public class UnixSocketStreamTransportMapping extends AbstractConnectionOrientedTransportMapping<UnixDomainAddress,
        UnixSocketStreamTransportMapping.SocketEntry> {

    private static final LogAdapter logger = LogFactory.getLogger(UnixSocketStreamTransportMapping.class);

    protected UnixDomainAddress listenAddress;
    protected ServerThread serverThread;

    public UnixSocketStreamTransportMapping(UnixDomainAddress listenAddress) {
        this.listenAddress = listenAddress;
    }

    public UnixSocketStreamTransportMapping() {
        this(null);
    }

    @Override
    public MessageLengthDecoder getMessageLengthDecoder() {
        return null;
    }

    @Override
    public void setMessageLengthDecoder(MessageLengthDecoder messageLengthDecoder) {
    }

    @Override
    public Class<? extends Address> getSupportedAddressClass() {
        return UnixDomainAddress.class;
    }

    @Override
    public Set<Class<? extends Address>> getSupportedAddressClasses() {
        return super.getSupportedAddressClasses();
    }

    @Override
    public UnixDomainAddress getListenAddress() {
        return listenAddress;
    }

    @Override
    public void sendMessage(UnixDomainAddress address, byte[] message, TransportStateReference tmStateReference,
                            long timeoutMillis, int maxRetries) throws IOException {
        if (getListenWorkerTask() == null || serverThread == null) {
//            if (isOpenSocketOnSending()) {
                listen();
//            }
//            else {
//                handleDroppedMessageToSend(address, message, tmStateReference, timeoutMillis, maxRetries);
//            }
        }
        if (serverThread != null) {
            if ((suspendedAddresses.size() > 0) && suspendedAddresses.contains(address)) {
                handleDroppedMessageToSend(address, message, tmStateReference, timeoutMillis, maxRetries);
            }
            else {
                serverThread.sendMessage(address, message, tmStateReference, sockets);
            }
        }
        /*
        SocketEntry socketEntry = sockets.get(address);
        if (socketEntry == null) {
            SocketChannel socketChannel = SocketChannel.open(StandardProtocolFamily.UNIX);
            socketChannel.connect(address.getSocketAddress());
            socketEntry = new SocketEntry(address, socketChannel);
            sockets.put(address, socketEntry);
        }
        ByteBuffer buffer = ByteBuffer.allocate(message.length);
        buffer.clear();
        buffer.put(message);
        buffer.flip();
        socketEntry.getSocketChannel().write(buffer);
         */
    }

    @Override
    public void close() throws IOException {
        if (listenAddress != null) {
            super.close();
            if (listenAddress.getPath() != null) {
                Files.deleteIfExists(listenAddress.getPath());
            }
        }
    }

    @Override
    public boolean close(UnixDomainAddress remoteAddress) throws IOException {
        AbstractSocketEntry<UnixDomainAddress> remoteSocketEntry = sockets.remove(remoteAddress);
        if (remoteSocketEntry == null) {
            return false;
        }
        remoteSocketEntry.getSocketChannel().close();
        return true;
    }

    @Override
    public void listen() throws IOException {
        if (getListenWorkerTask() != null) {
            throw new SocketException("Port already listening");
        }
        serverThread = new ServerThread(listenAddress);
        if (logger.isInfoEnabled()) {
            logger.info("Unix socket address " + getListenAddress() + " bound successfully");
        }
        listenWorkerTask = SNMP4JSettings.getThreadFactory().createWorkerThread(
                "UnixSocketStreamTransportMapping_" + getListenAddress(), serverThread, true);
        if (getConnectionTimeout() > 0) {
            // run as daemon
            socketCleaner = SNMP4JSettings.getTimerFactory().createTimer();
        }
        getListenWorkerTask().run();
    }

    @Override
    public boolean isListening() {
        return serverThread != null && !serverThread.isStop();
    }

    @Override
    public void wakeupServerSelector() {
        serverThread.getSelector().wakeup();
    }

    private static void postBind(UnixDomainAddress address) {
        address.getPath().toFile().deleteOnExit();
    }

    public static class SocketEntry extends AbstractSocketEntry<UnixDomainAddress> {

        public SocketEntry(UnixDomainAddress address, SocketChannel socket) {
            super(address, socket);
            // with Unix Domain sockets there is no handshaking, thus we are finished if connected
            setHandshakeFinished(true);
        }

        @Override
        public String toString() {
            return getPeerAddress() + getSocketChannel().toString();
        }

        @Override
        public void closeSession() {
            try {
                socketChannel.close();
            } catch (IOException e) {
                throw new UncheckedIOException(e);
            }
        }

        @Override
        public Object getSessionID() {
            return socketChannel;
        }
    }

    /**
     * Read the next network packet for the application.
     *
     * @param socketChannel
     *         the socket channel to read the data from.
     * @return a byte buffer with the application data of the packet.
     * @throws IOException
     *         if there occurs an IO exception during preparation.
     * @since 3.7
     */
    protected Optional<ByteBuffer> readNextInPacket(SocketChannel socketChannel)
            throws IOException {
        ByteBuffer buf = ByteBuffer.allocate(getMaxInboundMessageSize());
        int bytesRead = socketChannel.read(buf);
        if (bytesRead < 0) {
            return Optional.empty();
        }
        byte[] bytes = new byte[bytesRead];
        buf.flip();
        buf.get(bytes);
        return Optional.of(ByteBuffer.wrap(bytes));
    }

    protected class ServerThread extends AbstractTransportServerThread<UnixDomainAddress, SocketEntry> {

        private ServerThread(UnixDomainAddress unixDomainAddress) throws IOException {
            super(UnixSocketStreamTransportMapping.this, unixDomainAddress);
            selector = Selector.open();
            initListener(unixDomainAddress, selector);
        }

        public Throwable getLastError() {
            return lastError;
        }

        public boolean isStop() {
            return stop;
        }

        public void setStop(boolean stop) {
            this.stop = stop;
        }

        protected boolean initListener(UnixDomainAddress address, Selector selector) throws IOException {
            if (address != null) {
                ProtocolFamily family = address.getFamily();
                SocketAddress socketAddress = address.getSocketAddress();
                if (socketAddress != null) {
                    ssc = ServerSocketChannel.open(family);
                    ssc.bind(socketAddress);
                    ssc.configureBlocking(false);
                    postBind(address);
                    ssc.register(selector, SelectionKey.OP_ACCEPT, null);
                    logger.info("Unix Domain Server listening on " + ssc.getLocalAddress());
                    return true;
                }
            }
            return false;
        }

        @Override
        public void run() {
            doServer(sockets);
/*
            try {
                while (!stop) {
                    selector.select();
                    var keys = selector.selectedKeys();
                    for (SelectionKey key : keys) {
                        SelectableChannel c = key.channel();
                        if (c instanceof ServerSocketChannel server) {
                            var ch = server.accept();
                            ch.configureBlocking(false);
                            if (logger.isInfoEnabled()) {
                                logger.info("UnixDomainServer: new connection from " + ch.getRemoteAddress() + " with ID "+
                                        nextConnectionId);
                            }
                            ch.register(selector, SelectionKey.OP_READ, nextConnectionId++);
                        } else {
                            SocketChannel socketChannel = (SocketChannel) c;
                            Socket socketCopy = socketChannel.socket();
                            int id = (Integer)key.attachment();
                            TransportStateReference tmStateReference =
                                    new TransportStateReference(UnixSocketStreamTransportMapping.this,
                                            listenAddress, null,
                                            SecurityLevel.undefined, SecurityLevel.undefined,
                                            false, socketCopy);
                            var byteBufferRead = readNextInPacket(socketChannel);
                            byteBufferRead.ifPresent(message ->
                                            fireProcessMessage(new UnixDomainAddress((java.net.UnixDomainSocketAddress)
                                                    socketCopy.getRemoteSocketAddress()), message, tmStateReference));
                            if (byteBufferRead.isEmpty()) {
                                String remote = socketChannel.getRemoteAddress().toString();
                                if (logger.isInfoEnabled()) {
                                    logger.info("UnixDomainServer: closing connection from: " + remote + " with ID "+ id);
                                }
                                socketChannel.close();
                            }
                        }
                    }
                }
                serverSocketChannel.close();
            } catch (IOException iox) {
                lastError = iox;
                logger.warn(iox);
                if (logger.isDebugEnabled()) {
                    iox.printStackTrace();
                }
                if (SNMP4JSettings.isForwardRuntimeExceptions()) {
                    throw new RuntimeException(iox);
                }
            }
            finally {
                try {
                    serverSocketChannel.close();
                } catch (IOException iox) {
                    lastError = iox;
                    logger.error("UnixDomainServer: server socker channel closing failed: "+iox.getMessage());
                }
            }
            */
        }

        @Override
        protected boolean readMessage(SelectionKey sk, SocketChannel socketChannel,
                                      UnixDomainAddress incomingAddress, SocketEntry socketEntry) throws IOException {
            TransportStateReference tmStateReference =
                    new TransportStateReference(UnixSocketStreamTransportMapping.this,
                            listenAddress, null,
                            SecurityLevel.undefined, SecurityLevel.undefined,
                            false, socketChannel);
            var byteBufferRead = readNextInPacket(socketChannel);
            byteBufferRead.ifPresent(message ->
                    fireProcessMessage(new UnixDomainAddress((java.net.UnixDomainSocketAddress)
                            incomingAddress.getSocketAddress()), message, tmStateReference));
            if (byteBufferRead.isEmpty()) {
                String remote = socketChannel.getRemoteAddress().toString();
                if (logger.isInfoEnabled()) {
                    logger.info("UnixDomainServer: closing connection from: " + remote + " with ID "+ socketEntry);
                }
                socketChannel.close();
                return false;
            }
            return true;
        }

        @Override
        protected void processQueues() {
            // nothing to do
        }

        @Override
        public SocketEntry createSocketEntry(UnixDomainAddress address, SocketChannel socketChannel,
                                             boolean useClientMode, TransportStateReference tmStateReference) {
            return new SocketEntry(address, socketChannel);
        }

        @Override
        protected SocketChannel openSocketChannel(ProtocolFamily family) throws IOException {
            return SocketChannel.open(family);
        }

        @Override
        public SocketEntry removeSocketEntry(UnixDomainAddress incomingAddress) {
            return sockets.remove(incomingAddress);
        }

        @Override
        protected UnixDomainAddress createIncomingAddress(SocketChannel socketChannel) throws IOException {
            return new UnixDomainAddress((UnixDomainSocketAddress) socketChannel.getRemoteAddress());
        }
    }
}
